@@ -1,5 +1,5 @@
 from sys import platform
-from numpy import deprecate
+import numpy as np
 import pandas as pd
 import os
 from simple_salesforce import Salesforce
@@ -22,11 +22,15 @@ sys.path.append("..")
 try:
     from custom_exceptions.salesforce_exceptions import *
     from utilities.df_utils import *
+    from utilities.async_utils import *
     from decorators import *
 except:
     from salesforce.custom_exceptions.salesforce_exceptions import *
     from salesforce.utilities.df_utils import *
+    from salesforce.utilities.async_utils import *
     from salesforce.decorators import *
+import asyncio
+from functools import wraps, partial
 
 
 
@@ -165,6 +169,10 @@ class Sf:
                     'sfMeta_is_connected':sfMeta.is_connected(), 
                     'bulk_api_sessionId':bulk.sessionId}
 
+    ### ----------------------------------------------------------------------------------------------------
+    ### Synchronous Data Functions
+    ### ----------------------------------------------------------------------------------------------------
+
     def query(self, query: str, excludedFields: Optional[List] = []) -> pd.DataFrame:
         """
         Using SFDC SOQL Syntax, and allowing for Relationships and group bys. 
@@ -279,7 +287,8 @@ class Sf:
                     result[list(x.keys())[0]] = result[list(x.keys())[0].split(".")[0]].\
                         apply(lambda z: z[list(x.values())[0][1]])
                     columns_to_remove.add(list(x.values())[0][0])
-            result.drop(columns_to_remove, axis=1, inplace=True)        
+            result.drop(columns_to_remove, axis=1, inplace=True)
+            result.replace(np.nan, None, inplace=True)        
 
                 
         return result
@@ -308,7 +317,7 @@ class Sf:
         if (object_api is None) or (len(record_dataframe) == 0):
             raise RequiredValuesNotProvidedDuringCreate()
         else:
-            result = getattr(self.sf.bulk, object_api).insert(record_dataframe.to_dict('records'))
+            result = getattr(self.sf.bulk, object_api).insert(record_dataframe.to_dict('records'), batch_size=10000, use_serial=False)
             result = pd.DataFrame(result)
             result = unpack_column(result, "errors")
 
@@ -366,19 +375,30 @@ class Sf:
             RequiredValuesNotProvidedDuringUpdate
         """
 
-        field_metadata = self.field_describe([object_api])
+        field_metadata = self.field_describe([object_api],attributes=['name','type','updateable','compoundFieldName'])
 
         if ((object_api is None) or 
-        (len(record_dataframe) == 0) or 
-        record_dataframe.columns.__contains__('Id') == False or
-        # checks whether any of the columns in the dataframe passed in are not valid field API names
-        ~record_dataframe.columns.isin(field_metadata[object_api]).all()
+        (len(record_dataframe) == 0)):
+            raise RequiredValuesNotProvidedDuringUpdate(message="Object API and Record Dataframe are required, ensure your record dataframe has at least 1 row and 1 column with an Id column")
+        elif record_dataframe.columns.__contains__('Id') == False:
+            raise RequiredValuesNotProvidedDuringUpdate(message="Record Dataframe must have an Id column")
+        elif ~record_dataframe.columns.isin(field_metadata[object_api]).all():
+            # checks whether any of the columns in the dataframe passed in are not valid field API names
+            raise RequiredValuesNotProvidedDuringUpdate(message="One of the columns in the record dataframe is not a valid field for the object")
+        elif record_dataframe.columns.drop('Id').isin(field_metadata[field_metadata['Updateable'] == False]['Account']).any():
+            # checks whether any of the columns in the dataframe passed in are not updateable
+            non_updatable_fields = list(set(record_dataframe.columns) - set(field_metadata[field_metadata['Updateable'] == True]['Account']))
+            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are not updatable: {', '.join(non_updatable_fields)}")
+        elif record_dataframe.columns.isin(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')]['Account']).any():
+            # checks whether any of the columns in the dataframe passed in are compound fields
+            compound_fields = list(set(record_dataframe.columns) & set(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')]['Account']))
+            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are compound fields, which are not updatable: {', '.join(compound_fields)}")
         # checks whether an reference (lookup/master-detail) field is included in the source dataframe
         #  or ~record_dataframe.columns.isin(field_metadata[field_metadata['Type'] == 'reference'][object_api]).any()
-        ):
-            raise RequiredValuesNotProvidedDuringUpdate()
+            
         else:
-            result = getattr(self.sf.bulk, object_api).update(record_dataframe.to_dict('records'))
+            record_dataframe.replace(np.nan, None, inplace=True)
+            result = getattr(self.sf.bulk, object_api).update(record_dataframe.to_dict('records'), batch_size=10000, use_serial=False)
             result = pd.DataFrame(result)
             result = unpack_column(result, "errors")
 
@@ -410,18 +430,28 @@ class Sf:
             SalesforceMalformedRequest
         """
 
-        field_metadata = self.field_describe([object_api], ['name', 'type', 'length','externalId'])
-
+        field_metadata = self.field_describe([object_api], ['name', 'type', 'length','externalId','updateable','compoundFieldName'])
+            
+            
         if ((object_api is None) or 
-        (len(record_dataframe) == 0) or 
-        # record_dataframe.columns.__contains__('Id') == False or
-        # checks whether any of the columns in the dataframe passed in are not valid field API names
-        ~record_dataframe.columns.isin(field_metadata[object_api]).all()
-         or external_id_field_api==""
-        ):
-            raise RequiredValuesNotProvidedDuringUpsert()
+        (len(record_dataframe) == 0)):
+            raise RequiredValuesNotProvidedDuringUpdate(message="Object API and Record Dataframe are required, ensure your record dataframe has at least 1 row and 1 column with an Id column")
+        elif ~record_dataframe.columns.isin(field_metadata[object_api]).all():
+            # checks whether any of the columns in the dataframe passed in are not valid field API names
+            raise RequiredValuesNotProvidedDuringUpdate(message="One of the columns in the record dataframe is not a valid field for the object")
+        elif external_id_field_api == "":
+            raise RequiredValuesNotProvidedDuringUpsert(message="External ID field API is required")
+        elif record_dataframe.columns.drop('Id').isin(field_metadata[field_metadata['Updateable'] == False]['Account']).any():
+            # checks whether any of the columns in the dataframe passed in are not updateable
+            non_updatable_fields = list(set(record_dataframe.columns) - set(field_metadata[field_metadata['Updateable'] == True]['Account']))
+            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are not updatable: {', '.join(non_updatable_fields)}")
+        elif record_dataframe.columns.isin(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')]['Account']).any():
+            # checks whether any of the columns in the dataframe passed in are compound fields
+            compound_fields = list(set(record_dataframe.columns) & set(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')]['Account']))
+            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are compound fields, which are not updatable: {', '.join(compound_fields)}")
         else:
-            result = getattr(self.sf.bulk, object_api).upsert(record_dataframe.to_dict('records'), external_id_field_api)
+            record_dataframe.replace(np.nan, None, inplace=True)
+            result = getattr(self.sf.bulk, object_api).upsert(record_dataframe.to_dict('records'), external_id_field_api, batch_size=10000, use_serial=False)
             result = pd.DataFrame(result)
             result = unpack_column(result, "errors")
 
@@ -510,24 +540,68 @@ class Sf:
             print("Extracted " + objectName + " successfully!")
         return sfdf
 
-    def join(self, 
-                dataframe, 
-                rt_dataframe, 
-                left_on="", 
-                right_on="", 
-                new_columns=[], 
-                suffix = "_new"):
+    ### ----------------------------------------------------------------------------------------------------
+    ### Asynchronous Data Functions
+    ### ----------------------------------------------------------------------------------------------------
+
+    async def async_upsert(self, object_api, record_dataframe, external_id_field_api, batchsize=2000, *args, **kwargs):
+        async_upsert = async_wrap(self.upsert)
+        batches = len(record_dataframe) / batchsize
+        result_list = await asyncio.gather(*[async_upsert(object_api= object_api, record_dataframe=batch, external_id_field_api=external_id_field_api) for batch in np.array_split(record_dataframe, batches)])
+        return pd.concat(result_list).reset_index(drop=True)
+
+    async def async_update(self, object_api, record_dataframe, batchsize=2000, *args, **kwargs):
+        async_update = async_wrap(self.update)
+        batches = len(record_dataframe) / batchsize
+        result_list = await asyncio.gather(*[async_update(object_api= object_api, record_dataframe=batch) for batch in np.array_split(record_dataframe, batches)])
+        return pd.concat(result_list).reset_index(drop=True)
+
+    async def async_create(self, object_api, record_dataframe, batchsize=2000, *args, **kwargs):
+        async_create = async_wrap(self.create)
+        batches = len(record_dataframe) / batchsize
+        result_list = await asyncio.gather(*[async_create(object_api= object_api, record_dataframe=batch) for batch in np.array_split(record_dataframe, batches)])
+        return pd.concat(result_list).reset_index(drop=True)
+
+    async def async_delete(self, object_api, record_dataframe, batchsize=2000, *args, **kwargs):
+        async_delete = async_wrap(self.delete)
+        batches = len(record_dataframe) / batchsize
+        result_list = await asyncio.gather(*[async_delete(object_api= object_api, record_dataframe=batch) for batch in np.array_split(record_dataframe, batches)])
+        return pd.concat(result_list).reset_index(drop=True)
+
+    ### ----------------------------------------------------------------------------------------------------
+    ### Asynchronous Metadata Functions
+    ### ----------------------------------------------------------------------------------------------------
+    async def async_get_non_updatable_fields(self, object_name: str):
         """
-        Joins and appends the Name and DeveloperName columns of record type to the dataframe.
-        The dataframe must contain a column named "RecordTypeId" with the 18 digit SFID of the record type.
+        Returns a list of fields that cannot be updated.
         """
-        df_columns = dataframe.columns.to_list()
-        if right_on not in new_columns:
-            new_columns.insert(0,right_on)
-        dataframe = pd.merge(dataframe, rt_dataframe[new_columns], how = 'inner', 
-                                left_on = left_on,right_on = right_on, suffixes=('', suffix))
-#         dataframe.drop([col for col in dataframe.columns if 'drop' in col], axis=1, inplace=True)
-        return dataframe
+        async_field_describe =  async_wrap(self.field_describe)
+        field_metadata = await async_field_describe([object_name], attributes=['name','updateable'])
+        non_updatable_fields = list(field_metadata[field_metadata['Updateable'] == False][object_name])[1:]
+        return non_updatable_fields
+
+    async def async_get_compound_field_names(self, object_name: str, include_name_compound_fields: bool=False):
+        """
+        Returns a list of compound fields. By default "Name" compound field names are excluded.
+        To return a list of compound fields including "Name" fields, set include_name_compound_fields to True.
+
+        Compound field names are fields that are composed of multiple fields. These fields are not updatable directly via the API
+        and may cause errors if updated directly.
+        """
+        async_field_describe =  async_wrap(self.field_describe)
+        field_metadata = await async_field_describe([object_name], attributes=['name','compoundFieldName'])
+
+        if include_name_compound_fields:
+            compound_field_names = list(field_metadata[(~field_metadata['Compoundfieldname'].isnull())][object_name])
+        elif include_name_compound_fields==False:
+            compound_field_names = list(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')][object_name])
+        else:
+            raise ValueError('include_name_compound_fields must be a boolean.')
+        return compound_field_names
+
+    ### ----------------------------------------------------------------------------------------------------
+    ### Synchronous Metadata Functions
+    ### ----------------------------------------------------------------------------------------------------
 
     def object_describe(self, sobject_api_name: str, export_excel: bool=False):
         """
@@ -612,17 +686,23 @@ class Sf:
     def field_describe(self, objects: List = ['Account','Address_vod__c','Child_Account_vod__c'], 
     attributes: List = ['name','type','length']) -> pd.DataFrame:
         """
-        TODO:
+        Returns a dataframe of the field metadata for the specified objects and attributes.
+
+        Parameters
+        ----------
+        objects : List, optional
+            A list of objects to get field metadata for. The default is ['Account','Address_vod__c','Child_Account_vod__c'].
+        attributes : List, optional
+            A list of attributes to get field metadata for. The default is ['name','type','length'].
+            For a full list of attributes, see 
+            https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_calls_describesobjects_describesobjectresult.htm
+            A list of attributes can also be found by using the object_describe() method's field attribute.
+
+        Returns
+        -------
+        pd.DataFrame
+            A dataframe of the field metadata for the specified objects and attributes.
         
-        Description of what it does
-        
-        Description of arguments and data types
-        
-        Description of return values and data types
-        
-        Description of Errors raised
-        
-        Extra Notes and Examples of Usage
         """
         outputList = []
         columnNames = []
@@ -738,7 +818,10 @@ class Sf:
         getattr(self.sf.mdapi, metadata_type).rename(previous_name, new_name)
     
 
-
+    ### ----------------------------------------------------------------------------------------------------
+    ### Utility Functions
+    ### ----------------------------------------------------------------------------------------------------
+    
     def sf_api_call(self, action, parameters = {}, method = 'get', data = {}):
         """
         Helper function to make calls to Salesforce REST API.
@@ -764,4 +847,23 @@ class Sf:
                 return r.json()
         else:
             raise Exception('API error when calling %s : %s' % (r.url, r.content))
-    
\ No newline at end of file
+
+
+    def join(self, 
+                dataframe, 
+                rt_dataframe, 
+                left_on="", 
+                right_on="", 
+                new_columns=[], 
+                suffix = "_new"):
+        """
+        Joins and appends the Name and DeveloperName columns of record type to the dataframe.
+        The dataframe must contain a column named "RecordTypeId" with the 18 digit SFID of the record type.
+        """
+        df_columns = dataframe.columns.to_list()
+        if right_on not in new_columns:
+            new_columns.insert(0,right_on)
+        dataframe = pd.merge(dataframe, rt_dataframe[new_columns], how = 'inner', 
+                                left_on = left_on,right_on = right_on, suffixes=('', suffix))
+#         dataframe.drop([col for col in dataframe.columns if 'drop' in col], axis=1, inplace=True)
+        return dataframe
